from flask import Blueprint, request
from flask_restful import Api, Resource, abort
from marshmallow import ValidationError
from flask_jwt_extended import JWTManager, jwt_required, create_access_token, get_jwt_identity
from datetime import timedelta

from portal.rest import ProjectSchema, ProjectConfigSchema, SubmissionSchema,\
    delete_entity, write_entity, find_course, find_project, find_user
from portal.tools.logging import log
from portal.database.models import Project, Submission
from portal.service.service import can_create_submission

projects = Blueprint('projects', __name__, url_prefix='/courses/<string:cid>/projects')
projects_api = Api(projects)


project_schema = ProjectSchema()
projects_schema = ProjectSchema(many=True, only=('id', 'name'))
config_schema = ProjectConfigSchema()
# add user information to submission schema? How?
submissions_schema = SubmissionSchema(many=True, only=('id', 'parameters', 'state', 'note'))
submission_schema = SubmissionSchema()


class ProjectResource(Resource):
    def get(self, cid, pid):
        log.info(f"GET to {request.url}")
        course = find_course(cid)
        if not course:
            abort(404, message=f"Could not find project {pid} in course {cid}: course not found.")
        project = find_project(course, pid)
        if not project:
            abort(404, message=f"Could not find project {pid} in course {cid}: project not found.")

        return project_schema.dump(project)

    def delete(self, cid, pid):
        log.info(f"DELETE to {request.url}")
        course = find_course(cid)
        if not course:
            abort(404, message=f"Could not delete project {pid} in course {cid}: course not found.")
        project = find_project(course, pid)
        if not project:
            abort(404, message=f"Could not delete project {pid} in course {cid}: project not found.")

        delete_entity(project)
        return '', 204

    def put(self, cid, pid):
        log.info(f"PUT to {request.url}")
        course = find_course(cid)
        if not course:
            abort(404, message=f"Could not update project {pid} in course {cid}: course not found.")
        project = find_project(course, pid)
        if not project:
            abort(404, message=f"Could not update project {pid} in course {cid}: project not found.")

        json_data = request.get_json()
        if not json_data:
            abort(400, message='No data provided for project update.')
        try:
            data = project_schema.load(json_data)[0]
        except ValidationError as err:
            log.info(f"Validation failed on: {err.messages}")
            return abort(422, errors=err.messages)  # TODO: flash, process errors

        project.name = data['name']
        write_entity(project)
        log.debug(f"Updated project {pid} to {project}.")
        return '', 204


class ProjectsList(Resource):
    def get(self, cid):
        log.info(f"GET to {request.url}")
        course = find_course(cid)
        if not course:
            abort(404, message=f"Could not list projects in course {cid}: course not found.")

        return projects_schema.dump(course.projects)

    def post(self, cid):
        log.info(f"POST to {request.url}")
        course = find_course(cid)
        if not course:
            abort(404, message=f"Could not create project in course {cid}: course not found.")
        json_data = request.get_json()
        if not json_data:
            abort(400, message='No data provided for project creation.')

        try:
            data = project_schema.load(json_data)[0]
        except ValidationError as err:
            log.info(f"Validation failed on: {err.messages}")
            return abort(422, errors=err.messages)  # TODO: flash, process errors

        new_project = Project(course=course, name=data['name'], test_files_source=data['test_files_source'])
        write_entity(new_project)
        log.debug(f"Created project={new_project} for course {cid}")
        return project_schema.dump(new_project)[0], 201


class ProjectConfigResource(Resource):
    def get(self, cid, pid):
        log.info(f"GET to {request.url}")
        course = find_course(cid)
        if not course:
            abort(404, message=f"Could not get configuration of project {pid} in course {cid}: course not found.")
        project = find_project(course, pid)
        if not project:
            abort(404, message=f"Could not get configuration of project {pid} in course {cid}: project not found")

        return config_schema.dump(project.config)

    def put(self, cid, pid):
        log.info(f"PUT to {request.url}")
        course = find_course(cid)
        if not course:
            abort(404, message=f"Could not update configuration of project {pid} in course {cid}: course not found.")
        project = find_project(course, pid)
        if not project:
            abort(404, message=f"Could not update configuration of project {pid} in course {cid}: project not found.")

        json_data = request.get_json()
        if not json_data:
            abort(400, message='No data provided for project configuration update.')
        try:
            data = config_schema.load(json_data)[0]
        except ValidationError as err:
            log.info(f"Validation failed on: {err.messages}")
            return abort(422, errors=err.messages)  # TODO: flash, process errors

        project.set_config(data)  # TODO: check - missing **? in role permissions too
        log.debug(f"Updated configuration for project {pid} in course {cid} to {data}.")
        return '', 204


class ProjectSubmissions(Resource):
    @jwt_required
    def get(self, cid, pid):
        log.info(f"GET to {request.url}")
        course = find_course(cid)
        if not course:
            abort(404, message=f"Could not list submissions of project {pid} in course {cid}: course not found.")
        project = find_project(course, pid)
        if not project:
            abort(404, message=f"Could not list submissions of project {pid} in course {cid}: project not found.")
        # TODO: user=<user_selector>` - filters submissions by user
        return submissions_schema.dump(project.submissions)

    @jwt_required
    def post(self, cid, pid):
        log.info(f"POST to {request.url}")
        user = find_user(get_jwt_identity())
        if not user:
            abort(401, message="Invalid access token: user not found.")
        # authorization here

        course = find_course(cid)
        if not course:
            abort(404, message=f"Could not create a new submission for user {user.username} in project {pid} "
                               f"in course {cid}: course not found.")
        project = find_project(course, pid)
        if not project:
            abort(404, message=f"Could not create a new submission for user {user.username} in project {pid} "
                               f"in course {cid}: project not found.")

        # check if a new submission can be created by this user in this project
        # TODO: configure conflict_timeout in project
        if not can_create_submission(user, project, conflict_timeout=timedelta(minutes=30)):
            # maybe use 409, 403
            abort(429, message=f"New submission for user {user.username} denied: a submission for this user "
                               f"is already being processed.")

        json_data = request.get_json()
        if not json_data:
            abort(400, message='No data provided for submission creation.')
        try:
            data = config_schema.load(json_data)[0]
        except ValidationError as err:
            log.info(f"Validation failed on: {err.messages}")
            return abort(422, errors=err.messages)  # TODO: flash, process errors

        # TODO: store all configuration relevant for submission scheduling:
        # project.submissions_allowed from, to, archive; project.config.submission_scheduler_config, submission_params
        new_submission = Submission(user=user, project=project, parameters=project.config.submission_parameters,
                                    review=data['review'])
        write_entity(new_submission)
        # download relevant files into Storage
        # TODO storage

        return submission_schema.dump(new_submission)


class ProjectTestFiles(Resource):
    def get(self, cid, pid):
        # Gets test files (feature)
        return "Not implemented."


projects_api.add_resource(ProjectResource, '/<string:pid>')
projects_api.add_resource(ProjectsList, '/')
projects_api.add_resource(ProjectConfigResource, '/<string:pid>/config')
projects_api.add_resource(ProjectSubmissions, '/<string:pid>/submissions')
projects_api.add_resource(ProjectTestFiles, '/<string:pid>/files')

